# coding: utf-8
"""
Nel fare questo modulo ho incontrato alcuni problemi irrisolti:

1. NullBooleanField: non riesco ad annullare un campo che sia
   già True or False. Es.: transactional  => occore usare ''

2. JsonField: mi tocca aggiungere json.dumps quando creo e json.loads
   quando deserializzo. Non mi pare molto logico

3. Non riesco a propagare correttamente gli errori di sparkpost
   fino al client

"""

import logging

import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from django.utils import six
from django.conf import settings
from django.contrib.auth import get_user_model
User = get_user_model()

logger = logging.getLogger('thx.sparkpost')


class MailingApiError(Exception):
    pass


class ValidationError(Exception):
    pass


class Resource(object):
    key = ""

    def __init__(self, base_uri):
        self.base_uri = base_uri
        self.auth = HTTPBasicAuth(*settings.MAILING_REST_AUTH)

    def raise_for_status(self):
        """Raises stored :class:`HTTPError`, if one occurred."""
        ## REST stores error message in content!!!
        http_error_msg = ''
        if isinstance(self.response.reason, bytes):
            # We attempt to decode utf-8 first because some servers
            # choose to localize their reason strings. If the string
            # isn't utf-8, we fall back to iso-8859-1 for all other
            # encodings. (See PR #3538)
            try:
                reason = self.response.reason.decode('utf-8')
            except UnicodeDecodeError:
                reason = self.response.reason.decode('iso-8859-1')
        else:
            reason = self.response.reason

        if 400 <= self.response.status_code < 500:
            http_error_msg = u'%s Client Error: %s for url: %s' % (
                self.response.status_code, reason, self.uri)

        elif 500 <= self.response.status_code < 600:
            http_error_msg = u'%s Server Error: %s for url: %s' % (
                self.response.status_code, reason, self.uri)

        msg = '%s %s' % (http_error_msg, self.response.json())

        if http_error_msg:
            raise HTTPError(msg, response=self.response)

    @property
    def uri(self):
        return "%s/%s" % (self.base_uri, self.key)

    def request(self, method, uri, *args, **kwargs):
        self.response = getattr(requests, method)(uri, auth=self.auth, *args, **kwargs)
        self.raise_for_status()
        try:
            return self.response.json()
        except:
            return None


class Contact(Resource):
    key = 'contacts/'

    def get(self, email=None):
        uri = "{}{}/".format(self.uri, email)
        return self.request('get', uri)

    def add_from_user(self, user):
        data = {
            'name': user.get_full_name(),
            'email': user.email
        }
        return self.add_or_update(data)

    def add_or_update(self, data):
        """Add a contact if not already present

        :arg data: must be a dict with keys ``name`` and ``email``
                   if it contains ``address`` it must be a dict and
                   ``name`` and ``email``
                   are searched for in that nested dict.
                   Can also be just an email
        """
        uri = self.uri

        if isinstance(data, six.string_types):
            data = {'email': data}
        if 'address' in data:
            data = data['address']
        try:
            uri = "{}{}/".format(self.uri, data['email'])
            return self.request('patch', uri, json=data)
        except requests.exceptions.HTTPError as e:
            if self.response.status_code == 404:
                return self.request('post', self.uri, json=data)

    def delete(self, email):
        uri = "{}{}/".format(self.uri, email)
        return self.request('delete', uri)

    def patch(self, email, data):
        uri = "{}{}/".format(self.uri, email)
        return self.request('patch', uri, json=data)

    def add_to_group(self, contact, group):
        """Add the contact to a group

        :arg contact: email of user
        :arg group: group to add the contact to
        """
        if isinstance(contact, User):
            data = self.get(contact.email)
        else:
            data = self.get(contact)
        groups = data['groups']
        if group not in groups:
            groups += [group]
        return self.patch(data['email'], {'groups': groups})

    def remove_from_group(self, contact, group):
        """Remove contact from group

        :arg contact: email of user
        :arg group: group from which to revoke
        """
        if isinstance(contact, User):
            data = self.get(contact.email)
        else:
            data = self.get(contact)
        groups = data['groups']
        if group in groups:
            groups.pop(groups.index(group))
        return self.patch(data['email'], {'groups': groups})

    def add_to_category(self, contact, category):
        """Add the contact to a category

        :arg contact: email of user
        :arg category: group to add the contact to
        """
        if isinstance(contact, User):
            data = self.get(contact.email)
        else:
            data = self.get(contact)
        categories = data['categories']
        if category not in categories:
            categories += [category]
        return self.patch(data['email'], {'categories': categories})

    def remove_from_category(self, contact, category):
        """Remove contact from category

        :arg contact: email of user
        :arg category: group from which to revoke
        """
        if isinstance(contact, User):
            data = self.get(contact.email)
        else:
            data = self.get(contact)
        categories = data['categories']
        if category in categories:
            categories.pop(categories.index(category))
        return self.patch(data['email'], {'categories': categories})

    def subscribe(self, contact, recipient_list):
        if isinstance(contact, User):
            data = self.get(contact.email)
        else:
            data = self.get(contact)
        lists = data['lists']
        info = data['email'], recipient_list
        if recipient_list not in lists:
            lists += [recipient_list]
            logger.info("Subscription: '{}' subscribed to '{}'".format(*info))
            return self.patch(data['email'], {'lists': lists})
        logger.info("Subscription: '{}' already subscribed to '{}'".format(*info))
        return data

    def unsubscribe(self, contact, recipient_list):
        if isinstance(contact, User):
            data = self.get(contact.email)
        else:
            data = self.get(contact)
        lists = data['lists']
        info = data['email'], recipient_list
        if recipient_list in lists:
            lists.pop(lists.index(recipient_list))
            logger.info("Unsubscribe: '{}' unsubscribed from '{}'".format(*info))
            self.patch(data['email'], {'lists': lists})
        logger.info("Unsubscribe: '{}' already unsubscribed from '{}'".format(*info))
        return data

    def opened_by_transmission(self, email, **filters):
        """Get last email for contact with 'email'

        :arg kwargs: any data that should filter out the result
        """
        ## Send it
        querystring = "&".join(["{}={}".format(key, value) for key, value in filters.items()])
        uri = "{}{}/opened_by_transmission/?{}".format(self.uri, email, querystring)
        return self.request('get', uri)


class Transmission(Resource):
    """Manages transmissions to SparkPost.
    It's possibile to set MAILING_FAKE_TEST to prevent real sending. In this case the
    data that would be sent to SparkPost are appended in self.outbox (that is a list)

    It's meant to be used in tests and deletion of messages is supposed to be done by hand
        """
    key = 'transmissions/'

    def get(self, id=None):
        uri = "{}{}/".format(self.uri, id)
        return self.request('get', uri)

    def send_by_template(self, template_id, recipients, data):
        data.update({'template': template_id})
        return self.send(recipients, data)

    def create(self, data=None, recipients=None, **kwargs):
        """create a transmission

        :arg data: the json data defining the transmission
        :arg kwrags: any kwargs will overwrite data
        """
        if not (recipients or 'recipients' in data or 'recipient_list' in data
                or 'recipients_json' in data):
            raise ValidationError('Missing recipients')

        if recipients and 'recipients' in data:
            raise ValidationError('Recipients can be defined just once (appears also in data)')
        # while debugging it's nicer if data and recipients remain clean
        data = data and data.copy() or {}
        if recipients:
            data.update(self.parse_recipients(recipients or data['recipients']))
        data.update(kwargs)
        ## context is more django-friendly...
        if 'context' in data:
            data['substitution_data'] = data['context']
            del data['context']
        ## Checks
        logger.info("Transmission: created with name %s", data.get('name'))
        if settings.MAILING_FAKE_SEND:
            if not hasattr(self, 'outbox'):
                self.outbox = []
            self.outbox += [data]
            return
        return self.request('post', self.uri, json=data)

    def update(self, data=None, id=None, **kwargs):
        """Update a transmission. Uses ``PATCH`` so that a partial update is possible

        :arg data: the data we want to change

        """
        if id:
            uri = "{}{}/".format(self.uri, id)
        elif 'links' in data:
            uri = data['links']['self']
        data = data or {}
        data.update(kwargs)
        return self.request('patch', uri, json=data)

    def send(self, recipients=None, data=None, **kwargs):
        """Create a transmission and send it

        :arg recipients: recipients. May be

            * a single email
            * a ``list_id``
            * a user
            * a collection of users.
        :arg data: the json data of the transmission
        :arg kwargs: any data that should overwrite data
        """
        ## Create transmission obj
        resp = self.create(data=data, recipients=recipients)
        if settings.MAILING_FAKE_SEND:
            return
        id = resp['id']
        ## Send it
        uri = "{}{}/send/".format(self.uri, id)
        logger.info("Transmission: sent with name {}".format(data.get('name')))
        return self.request('post', uri)

    def get_address_json(self, user):
        return user.email
        # return {
        #     'address': {
        #         'name': user.get_full_name(),
        #         'email': user.username
        #     }
        # }

    def parse_recipients(self, recipients):
        """return a dict to update recipients

        a string with '@' is an email
        a string w/o '@' is a recipient list
        a User object
        an iterable can hold User objects or string objects.
        In the latter case prev rules are applied

        """
        data = {}
        if isinstance(recipients, six.string_types):
            if '@' in recipients:
                data['recipients'] = [recipients]
            else:
                data['recipient_list'] = recipients
        elif isinstance(recipients, User):
            data['recipients'] = [self.get_address_json(recipients)]
        elif hasattr(recipients, '__iter__'):
            data['recipients'] = [self.get_address_json(u)
                                  for u in recipients if isinstance(u, User)]
            data['recipients'] += [u for u in recipients if isinstance(u, six.string_types)]
        return data

    def send_by_id(self, id):
        uri = "{}{}/send/".format(self.uri, id)
        return self.request('post', uri)


class SentEmail(Resource):
    key = 'sentemail/'

    def get(self, id=None):
        uri = "{}{}/".format(self.uri, id)
        self.resp = self.request('get', uri)
        if self.resp.ok:
            return self.resp.json()

    def get_events(self, id):
        uri = "{}{}/".format(self.uri, id)
        self.resp = self.request('get', uri)
        if self.resp.ok:
            return self.resp.json()


class MailingClient(object):

    def __init__(self, base_uri=None):

        self.base_uri = base_uri or settings.MAILING_BASE_URI
        self.auth = HTTPBasicAuth(*settings.MAILING_REST_AUTH)
        self.contacts = Contact(self.base_uri)
        self.transmissions = Transmission(self.base_uri)

    def get(self, url='/'):
        url = self.base_uri + url
        self.resp = requests.get(url, auth=self.auth)
        if self.resp.ok:
            return self.resp.json()
        else:
            self.resp.raise_for_status()

    def test(self):
        return self.get()

    def post(self, url, data):
        url = self.base_uri + url
        self.resp = requests.post(url, json=data, auth=self.auth)
        if self.resp.ok:
            return self.resp.json()
        else:
            self.resp.raise_for_status()

    def delete(self, url, pk):
        url = self.base_uri + url + pk
        self.resp = requests.delete(url, auth=self.auth)
        self.resp.raise_for_status()

mailing_client = MailingClient()
