# coding: utf-8
"""
"""

import datetime
import logging
import json

from django.utils import six
from django.utils.encoding import python_2_unicode_compatible
from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.contrib.postgres.fields import JSONField
from django.contrib import messages

from jmb.core.db.models import DateModel, UserModel, StatusModel
from jmb.core.middleware.thread_local import get_request

from sparkpost import SparkPost


logger = logging.getLogger('thx.sparkpost')


class SparkPostValidationError(Exception):
    pass


class RecipientList(DateModel, UserModel, StatusModel):
    """https://developers.sparkpost.com/api/recipient-lists/
    """
    list_id = models.CharField(max_length=50, primary_key=True)
    name = models.CharField(max_length=50, unique=True)
    description = models.CharField(max_length=1024, null=True, blank=True)
    tags = JSONField(null=True, blank=True)
    metadata = JSONField(null=True, blank=True)
    substitution_data = JSONField(null=True, blank=True)
    last_sync = models.DateTimeField(null=True)
    total_accepted_recipients = models.IntegerField(null=True, blank=True)
    total_rejected_recipients = models.IntegerField(null=True, blank=True)

    class Meta(object):
        verbose_name = _('Recipient list')
        verbose_name_plural = _('Recipient lists')

    def create(self):
        """Create a list on sparkpost"""
        pass

    def update(self):
        """Update the list on sparkpost"""
        pass

    def get_recipients(self):
        ret = []
        for c in self.contacts.all():
            ret += [c.get_data()]
        return ret

    def get(cls, list_id=None):
        from tre.marketing import utils
        sp = utils.TreSparkPost()
        data = sp.templates.get(list_id, show_recipients=True)
        return data

    def spark_sync(self):
        """Update a sparkpost list based on our recipient-list"""
        sp = SparkPost(settings.SPARKPOST_API_KEY)
        kwargs = {
            'id': self.list_id,
            'name': self.name,
            'description': self.description,
            'recipients': self.get_recipients()
        }
        if self.last_sync:
            response = sp.recipient_lists.update(kwargs.pop('id'), **kwargs)
        else:
            response = sp.recipient_lists.create(**kwargs)

        self.total_rejected_recipients = response['total_rejected_recipients']
        self.total_accepted_recipients = response['total_accepted_recipients']
        self.last_sync = datetime.datetime.now()
        self.save()

    def spark_delete(self):
        """Delete a list that is present in sarkpost"""
        sp = SparkPost(settings.SPARKPOST_API_KEY)
        return sp.recipient_lists.delete(self.list_id)

    def __str__(self):
        return self.description or self.name


class Content(DateModel, UserModel, StatusModel):

    description = models.CharField(max_length=50)
    html = models.TextField(null=True, blank=True)
    text = models.TextField(null=False, blank=True)
    subject = models.CharField(max_length=120)
    from_name = models.CharField(max_length=100, null=True, blank=True)
    from_email = models.CharField(max_length=100)
    reply_to = models.CharField(max_length=50, null=True, blank=True)
    headers = models.CharField(max_length=150, null=True, blank=True)
    attachments = JSONField(null=True, blank=True)
    inline_images = JSONField(null=True, blank=True)

    def get_data(self):

        data = {
            'html': self.html,
            'text': self.text,
            'subject': self.subject,
            'from_email': {'name': self.from_name, 'email': self.from_email},
        }

        return data

    def __str__(self):
        return self.description


class Template(DateModel, UserModel, StatusModel):
    """Sparkpost template described here_


    .. _here: https://developers.sparkpost.com/api/templates/
    """

    template_id = models.CharField(max_length=64, null=True, blank=True)
    name = models.CharField(max_length=1024)
    description = models.CharField(max_length=1024)
    published = models.BooleanField()
    content = models.ForeignKey(Content, blank=True, null=True)
    pushed = models.DateTimeField(null=True, blank=True)
    transactional = models.BooleanField(default=False)
    click_tracking = models.BooleanField(default=False)
    open_tracking = models.BooleanField(default=False)

    class Meta:
        ordering = ('-id',)

    def __str__(self):
        return "{} - {}".format(self.template_id, self.name)

    def get_data(self):
        "Data used to sync the template remotely"

        data = {
            'id': self.template_id,
            'name': self.name,
            'description': self.description,
            'published': self.published,
            'options': {
                'transactional': self.transactional,
                'click_tracking': self.click_tracking,
                'open_tracking': self.open_tracking,
            }
        }
        data.update(self.content.get_data())

        return data

    def push(self):
        sp = SparkPost(settings.SPARKPOST_API_KEY)
        logger.info("pushing template {}".format(self.name))
        if self.pushed:
            sp.templates.update(self.template_id, **self.get_data())
        else:
            sp.templates.create(**self.get_data())
        self.pushed = datetime.datetime.now()
        self.save()

    @classmethod
    def get(cls, template_id=None):
        from tre.marketing import utils
        sp = utils.TreSparkPost()
        data = sp.templates.get(template_id)
        return data

    @classmethod
    def import_from_sparkpost(cls, template_id):
        data = cls.get(template_id)
        try:
            tpl = cls.objects.get(template_id=template_id)
            logger.info("Template already present: %s" % template_id)
        except cls.DoesNotExist:
            tpl = cls(template_id=data['id'])
            logger.info("Creating new Template: %s" % template_id)
        # FIXME: missing options
        tpl.name = data['name']
        tpl.description = data['description']
        tpl.published = data['published']
        if 'transactional' in data['options']:
            tpl.transactional = data['options']['transactional']
        if 'open_tracking' in data['options']:
            tpl.open_tracking = data['options']['open_tracking']
        if 'click_tracking' in data['options']:
            tpl.click_tracking = data['options']['click_tracking']
        content = Content()
        content.description = data['description']
        if 'text' in data['content']:
            content.text = data['content']['text']
        if 'html' in data['content']:
            content.html = data['content']['html']
        content.subject = data['content']['subject']
        if 'reply_to' in data['content']:
            content.reply_to = data['content']['reply_to']
        if 'headers' in data['content']:
            content.headers = data['content']['headers']
        if isinstance(data['content']['from'], six.string_types):
            content.from_email = data['content']['from']
        elif isinstance(data['content']['from'], dict):
            if 'name' in data['content']['from']:
                content.from_name = data['content']['from']['name']
            content.from_email = data['content']['from']['email']
        content.save()
        tpl.content = content
        tpl.save()


class Transmission(DateModel, UserModel, StatusModel):
    """Sparkpost action described here_


    .. _here: https://developers.sparkpost.com/api/transmissions/

    """

    transmission_id = models.CharField(max_length=50, null=True, blank=True)
    name = models.CharField(max_length=150, null=True, blank=True)
    state = models.CharField(max_length=15, null=True, blank=True)
    recipient_list = models.ForeignKey(RecipientList, null=True, blank=True)
    recipients_json = JSONField(null=True, blank=True)
    campaign = models.ForeignKey(Campaign, null=True, blank=True)
    description = models.CharField(max_length=80, null=True, blank=True)
    metadata = JSONField(null=True, blank=True)
    substitution_data = JSONField(null=True, blank=True)
    return_path = models.CharField(max_length=50, null=True, blank=True)
    total_recipients = models.IntegerField(null=True, blank=True)
    num_generated = models.IntegerField(null=True, blank=True)
    num_failed_generation = models.IntegerField(null=True, blank=True)
    num_invalid_recipients = models.IntegerField(null=True, blank=True)
    start_time = models.DateTimeField(null=True, blank=True)
    open_tracking = models.NullBooleanField()
    click_tracking = models.NullBooleanField()
    transactional = models.NullBooleanField()
    sandbox = models.BooleanField(default=False)
    inline_css = models.BooleanField(default=False)
    transmitted = models.DateTimeField(null=True, blank=True)
    reply_to = models.EmailField('ReplyTo', null=True, blank=True)

    template = models.ForeignKey(Template, blank=True, null=True)
    content = models.ForeignKey(Content, blank=True, null=True)
    recipients = models.ManyToManyField(
        Contact, related_name='transmissions', blank=True)

    class Meta(object):
        verbose_name = _('Transmission')
        verbose_name_plural = _('Transmissions')

    @property
    def options(self):
        opts = {}
        for attr in ('start_time', 'open_tracking', 'click_tracking', 'transactional',
                     'sandbox', 'inline_css'):
            opts[attr] = getattr(self, attr)
        return opts

    def __str__(self):
        return self.name or self.transmission_id or ''

    def get_options(self):
        opts = {
            'substitution_data': self.substitution_data,
            # 2017-02-11T08:00:00-04:00
            'start_time': self.start_time and self.start_time.strftime('%Y-%m-%dT%H:%M:00+02:00'),
            'transactional': self.transactional,
            'inline_css': self.inline_css,
            # python libreary differs from SPARK offcial APIs...
            'track_opens': self.open_tracking,
            'track_clicks': self.click_tracking,
            # 'open_tracking': self.open_tracking,
            # 'click_tracking': self.click_tracking,
        }
        return opts

    def get_recipients(self):
        if self.recipients_json:
            return {'recipients': self.recipients_json}
        if self.recipient_list:
            return {'recipient_list': self.recipient_list.list_id}
        recipients = [r.get_data() for r in self.recipients.filter(status__gte=1)]
        return {'recipients': recipients}

    def get_recipients_as_contacts(self):
        if self.recipient_list:
            return self.recipient_list.contacts.all()
        if self.recipients_json:
            email_list = [x['address']['email'] for x in self.recipients_json]
            contacts = Contact.objects.filter(email__in=email_list)
            if len(email_list) < contacts.count():
                msg = 'Transmission %s has fewer contacts that recipient in recipients_json'
                logger.warning(msg, self.transmission_id)
            return contacts
        return self.recipients.all()

    def get_data(self):
        data = self.get_options()
        data.update(self.get_recipients())  # recipients or recipient_list
        if self.campaign:
            data['campaign'] = self.campaign.name
        if self.reply_to:
            data['reply_to'] = self.reply_to
        if self.template:
            data['template'] = self.template.template_id
            if self.transactional is None:
                data['transactional'] = self.template.transactional
            if self.click_tracking is None:
                data['click_tracking'] = self.template.click_tracking
            if self.open_tracking is None:
                data['open_tracking'] = self.template.open_tracking
        else:
            data.update(self.content.get_data())

        return {key: val for key, val in data.items() if val not in ('', None)}

    def send(self, force=False):
        """Send the transmission to Sparkpost and update record with
        returned info. Abort if already sent.

        :arg force: force a retransmission (debug)
        """
        if not self.status:
            raise Exception("Won't send a disabled transmission")
        if self.transmission_id and not force:
            request = get_request()
            msg = _("This transmission has already been sent, create a copy first")
            if request:
                messages.error(request, msg)
            else:
                logger.error(msg)
            raise SparkPostValidationError(msg)

        sp = SparkPost(settings.SPARKPOST_API_KEY)
        logger.info("sending transmission {}".format(self.id))
        self.transmitted = datetime.datetime.now()
        response = sp.transmission.send(**self.get_data())

        self.transmission_id = response['id']
        self.total_recipients = response['total_accepted_recipients']
        self.total_invalid_recipients = response['total_rejected_recipients']
        self.save()
        self.generate_sent_email()
        return response

    def generate_sent_email(self):
        """Create a SentEmail object used to collect events
        """
        email_list = []
        for c in self.get_recipients_as_contacts():
            email_list.append(SentEmail(contact=c, transmission=self, date_create=self.date_create))
        SentEmail.objects.bulk_create(email_list)

    def get_events(self, filters=''):

        from tre.marketing import utils
        sp = utils.TreSparkPost()
        data = sp.events.get_by_transmission(self.transmission_id, filters=filters)
        return data

    def add_events(self, filters=''):
        """Get all events for transmission and create all event objects"""

        data = self.get_events(filters=filters)
        j = 0
        for d in data:
            Event.add_event_from_json(d)
            j += 1
        return j

    def get(self):
        from tre.marketing import utils
        sp = utils.TreSparkPost()
        data = sp.transmission.get(self.transmission_id)
        return data

    def preview(self, output=None, draft=False, ):
        """Return a preview of this transmission using the same context

        :arg output: if preset must be the path of a file to be written
        :arg draft: if True, the draft version is used
        """
        from tre.marketing import utils
        sp = utils.TreSparkPost()

        data = self.substitution_data
        tpl = self.template.template_id
        res = sp.templates.preview(template_id=tpl, substitution_data=data, draft=draft)
        if output:
            with open(output, 'w') as f:
                f.write(res['html'])
            return
        return res


class SentEmail(DateModel, UserModel, StatusModel):

    """What was really sent, should be possible to rebuild, based on transmission"""

    user = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True)
    contact = models.ForeignKey(Contact, null=True, blank=True, related_name='sent_emails')
    transmission = models.ForeignKey(Transmission, related_name='emails')
    context = JSONField(null=True, blank=True)
    injected = models.DateTimeField(null=True, blank=True)
    bounced = models.DateTimeField(null=True, blank=True)
    delivered = models.DateTimeField(null=True, blank=True)
    opened = models.DateTimeField(null=True, blank=True)
    failed = models.DateTimeField(null=True, blank=True)
    clicked = models.DateTimeField(null=True, blank=True)
    unsubscribe = models.DateTimeField(null=True, blank=True)
    registered = models.DateTimeField(null=True, blank=True)

    class Meta(object):
        verbose_name = _('Sent Email')
        verbose_name_plural = _('Sent Emails')

    def get_content(self):
        """Rebuild the mail actually sent from template and context"""
        pass

    def preview(self, output=None, draft=False, ):
        """Return a preview of this transmission using the same context
        """

        from tre.marketing import utils
        sp = utils.TreSparkPost()

        data = self.transmission.substitution_data
        tpl = self.transmission.template.template_id
        res = sp.templates.preview(template_id=tpl, substitution_data=data, draft=draft)
        if output:
            with open(output, 'w') as f:
                f.write(res['html'])
            return
        return res

    def __str__(self):
        return '{x.contact.email} {x.date_create}'.format(x=self)


class Event(models.Model):

    """Events collected from saprkpost (https://developers.sparkpost.com/api/events/)
    """
    EVENT_CHOICES = (
        ('delivery', _('delivery')),
        ('injection', _('injection')),
        ('bounce', _('bounce')),
        ('delay', _('delay')),
        ('policy_rejection', _('policy')),
        ('out_of_band', _('out')),
        ('open', _('open')),
        ('click', _('click')),
        ('generation', _('generation')),
        ('failure', _('failure')),
        ('generation_rejection', _('generation')),
        ('spam_complaint', _('spam')),
        ('list_unsubscribe', _('list')),
        ('link_unsubscribe', _('link')),
    )
    timestamp = models.DateTimeField(null=True)
    email = models.ForeignKey(SentEmail, related_name='events')
    event_type = models.CharField(max_length=20, choices=EVENT_CHOICES)
    link = models.URLField(null=True, blank=True)
    data = JSONField()
    event_id = models.CharField(max_length=30, db_index=True, null=True, blank=True)
    transmission_id = models.CharField(max_length=30, db_index=True, null=True, blank=True)

    class Meta(object):
        verbose_name = _('Event')
        verbose_name_plural = _('Events')

    @classmethod
    def add_event_from_json(cls, data, fake=False, upsert=False):
        """Add event from the json arriving from sparkpost

        :arg data: the json data as received from sparkhost
        :arg fake: if True, the events will not be saved (Test purpose)
        :arg upsert: accept
        """
        transmission_id = data['transmission_id']
        recipient_email = data['rcpt_to']
        event_id = data['event_id']

        try:
            timestamp = datetime.datetime.fromtimestamp(int(data['timestamp']))
        except (AttributeError, TypeError, ValueError):
            timestamp = parse(data['timestamp'])
            #timestamp = datetime.datetime.strptime(timestamp_str, '%Y-%m-%dT%H:%M:%S')
        # there may be more than one email...
        ref = SentEmail.objects.filter(
            transmission__transmission_id=transmission_id,
            contact__email=recipient_email,
        ).first()  # NOTE: disregarding possible repetition of an address in list
        if not ref:
            logger.warning("Couldn't find transmission {id} for {email}".format(
                id=transmission_id, email=recipient_email))
            return

        event_type = data['type']

        new = {
            'email': ref,
            'timestamp': timestamp,
            'event_type': event_type,
            'data': data,
            'event_id': event_id,
            'transmission_id': transmission_id
        }
        if event_type == 'click':
            new['link'] = data['target_link_url']
        if event_type == 'injection':
            ref.injected = timestamp
        elif event_type == 'click':
            ref.clicked = timestamp
        elif event_type == 'bounce':
            ref.bounced = timestamp
        elif event_type == 'open':
            ref.opened = timestamp
        elif event_type == 'failure':
            ref.failed = timestamp
        elif event_type in ('list_unsubscribe', 'link_unsubscribe'):
            ref.unsubscribed = timestamp
        elif event_type == 'delivery':
            ref.delivered = timestamp
        if fake:
            return cls(**new)
        ref.save()
        if upsert:
            obj, created = cls.objects.get_or_create(event_id=event_id, defaults=new)
        else:
            obj = cls.objects.create(**new)

        return obj

    def __str__(self):
        return "{} {}".format(self.email, self.event_type)


