# coding: utf-8

from __future__ import unicode_literals

from . import models

from rest_framework import serializers
from rest_framework.reverse import reverse


class EventSerializer(serializers.HyperlinkedModelSerializer):

    type = serializers.CharField(source='event_type')
    transmission_id = serializers.CharField(source='transmission')

    class Meta:
        model = models.Event
        fields = ('type', 'link', 'transmission_id')

    def create(self, validated_data):
        return models.Event.add_event_from_json(self.context['request'].data)


class RecipientListSerializer(serializers.HyperlinkedModelSerializer):

    links = serializers.SerializerMethodField()

    class Meta:
        model = models.RecipientList
        #fields = ('list_id', 'name', 'description', 'links', 'contacts')
        fields = ('list_id', 'name', 'description', 'links', )
        lookup_field = 'name'
        extra_kwargs = {
            'contacts': {'lookup_field': 'email'}
        }

    def get_links(self, obj):
        request = self.context['request']
        return {
            'self': reverse('recipientlist-detail',
                            kwargs={'pk': obj.pk}, request=request),
            # 'contacts': reverse('contact-list',
            #                     kwargs={'name': obj.name}, request=request),
        }


class TransmissionSerializer(serializers.HyperlinkedModelSerializer):

    links = serializers.SerializerMethodField()
    template = serializers.SlugRelatedField(
        slug_field='template_id', allow_null=True, required=False, read_only=False,
        queryset=models.Template.objects.all()
    )
    recipient_list = serializers.SlugRelatedField(
        slug_field='list_id', allow_null=True, required=False, read_only=False,
        queryset=models.RecipientList.objects.all()
    )
    recipients = serializers.SlugRelatedField(
        slug_field='email', allow_null=True, required=False, read_only=False,
        queryset=models.Contact.objects.all(), many=True,
    )

    class Meta:
        model = models.Transmission
        fields = (
            'id', 'links', 'transmission_id', 'name', 'state',
            'description', 'substitution_data', 'return_path',
            'open_tracking', 'click_tracking', 'transactional',
            'template', 'recipient_list', 'recipients_json', 'recipients'
        )
        lookup_field = 'transmission_id'
        extra_kwargs = {
            'recipients': {'lookup_field': 'email'},
            'template': {'lookup_field': 'template_id'},
        }

    def get_links(self, obj):
        request = self.context['request']
        return {
            'self': reverse('transmission-detail',
                            kwargs={'pk': obj.pk}, request=request),
        }


class TemplateSerializer(serializers.HyperlinkedModelSerializer):

    links = serializers.SerializerMethodField()

    class Meta:
        model = models.Template
        fields = ('template_id', 'name', 'description', 'links')
        lookup_field = 'name'

    def get_links(self, obj):
        request = self.context['request']
        return {
            'self': reverse('template-detail',
                            kwargs={'template_id': obj.template_id}, request=request),
        }


class GroupSerializer(serializers.HyperlinkedModelSerializer):

    links = serializers.SerializerMethodField()

    class Meta:
        model = models.Group
        fields = ('name', 'description', 'links')
        lookup_field = 'name'

    def get_links(self, obj):
        request = self.context['request']
        return {
            'self': reverse('group-detail',
                            kwargs={'name': obj.name}, request=request),
        }


class ContactSerializer(serializers.HyperlinkedModelSerializer):
    lists = serializers.PrimaryKeyRelatedField(
        many=True, queryset=models.RecipientList.objects.all(), required=False)
    groups = serializers.SlugRelatedField(
        slug_field='name', allow_null=True, many=True, required=False,
        queryset=models.Group.objects.all()
    )
    categories = serializers.SlugRelatedField(
        slug_field='key', allow_null=True, many=True, required=False,
        queryset=models.Category.objects.all()
    )
    links = serializers.SerializerMethodField()

    class Meta:
        model = models.Contact
        fields = ('url', 'email', 'name', 'lists', 'categories', 'groups', 'links')
        lookup_field = 'email'

    def get_links(self, obj):
        request = self.context['request']
        return {
            'self': reverse('contact-detail',
                            kwargs={'email': obj.email}, request=request),
        }
